"""The `star_ray_xml` package is an extension of [`star_ray`](https://github.com/dicelab-rhul/star-ray) that uses XML as the description language for the state of an environment. It supports read/write queries of the XML state via [xpath](https://www.w3schools.com/xml/xpath_intro.asp), providing a language with which agents can take actions and observe their environment.

Important classes:
    `XMLState` : which defines the public API which an `Ambient` may use to access the underlying state.
    `_XMLState` : the default (internal) implementation of `XMLState` that is backed by the well-known `lxml` package.
    `XMLAmbient` : the default implementation of an `Ambient` (see `star_ray` package) that makes use of XML as its state description language. It exposes the standard `__update__`, `__select__` API and is read and mutated via `XMLQuery` events (see below).

Query classes:
    Select : Read-only query that selects (retrieves) elements and their attributes from the XML state.
    Update : Write query that will update element attributes.
    Replace : Write query that will replace entire elements.
    Delete : Write query that will delete elements or their attributes.
    Insert : Write query that will insert new elements.
    XMLQuery : The base class for all queries, defines the `__execute__` method which is the method that effectively defines how the query mutates (or reads) the state. It provides direct access to the `XMLState` API and may be subclassed to provide more user-friendly queries, especially where the operation may require access to various XML attributes which might otherwise require additional queries (these instead can be read or written to directly).

See the documentation in each class for details on their use. These queries can also be constructed using the following factory methods: [`select`, `update`, `replace`, `delete`, `insert`] which provide some conveniences.
"""

from .query import (
    select,
    insert,
    delete,
    replace,
    update,
    Expr,
    Select,
    Insert,
    Delete,
    Replace,
    Update,
    XMLQuery,
    XMLUpdateQuery,
    XPathQuery,
    XMLQueryError,
)
from .state import XMLState, _XMLState
from .ambient import XMLAmbient
from .sensor import XMLSensor

__all__ = (
    "XMLAmbient",
    "XMLState",
    "_XMLState",
    "XMLSensor",
    "select",
    "insert",
    "delete",
    "replace",
    "update",
    "Select",
    "Insert",
    "Delete",
    "Replace",
    "Update",
    "Expr",
    "Expr",
    "XMLQuery",
    "XMLUpdateQuery",  # TODO others coming
    "XPathQuery",
    "XMLQueryError",
)



"""Some utility functions for querying XML."""

import ast
import html
from lxml import etree as ET


def xml_to_primitive(value: str):
    """Converts a string attribute value to an appropriate Python type using ast.literal_eval."""
    try:
        # Safely evaluate value as a Python literal
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        # Return the original value if it's not a valid Python literal
        return value


def xml_element_to_string(
    element: ET.ElementBase, unescape: bool = True, with_tail: bool = False
):
    """Converts an lxml Element into a string."""
    result = ET.tostring(
        element,
        encoding=str,
        with_tail=with_tail,
    )
    if unescape:
        return html.unescape(result)
    else:
        return result



"""Defines the `XMLSensor` class which is a useful sensor implementation for observing XML related data."""

from star_ray.agent import Agent, Sensor, attempt
from star_ray.pubsub import Subscribe
from .query import select, Select, XMLQuery


class XMLSensor(Sensor):
    """Sensor that will observe all changes to an XMLAmbient. It subscribes to receive all events that subclass `XMLQuery` and initially attempts to sense all XML data (as XML source code)."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        self._subscriptions = (XMLQuery,)

    @attempt
    def select_all(self) -> Select:
        """An attempt method that takes an action to select all current xml data.

        Returns:
            Select: the action
        """
        action = select("/*")  # select all xml data
        return action

    @attempt
    def element_exists(self, element_id: str) -> Select:
        """An attempt method that selects the `id` attribute from a given element. The `id` will match `element_id` if it is found. This can be used to check whether an element exists (if resulting observation is non-empty).

        Args:
            element_id (str): the elemnet to check.

        Returns:
            Select: the action
        """
        return select(f"//*[@id='{element_id}']", ["id"])

    def on_add(self, agent: Agent) -> None:  # noqa: D102
        super().on_add(agent)
        # initially get all xml data - this will be avaliable on the first sense cycle
        self.select_all()

    def __subscribe__(self) -> list[Subscribe]:  # noqa: D105
        return [Subscribe(topic=sub) for sub in self._subscriptions]



import re
import ast
from typing import Any
from lxml import etree as ET
from .query import XMLQueryError, Expr

XML_START_PATTERN = re.compile(r"^\s*<")


class _Element:
    """Wraps an `lxml` element (which may be full XML elements, their attributes, literals (e.g. text) or other properties) adding some additional functionality for use in implementations of `XMLState`. `XMLState` will handle the creation of instances of `_Element`, it is not part of the public API but is stable and documented regardless."""

    def __init__(self, base: ET.ElementBase):
        super().__init__()
        self._base = base

    def get_root(self) -> "_Element":
        """Retrieves the root element of the tree containing this element.

        Returns:
            _Element: The root element.
        """
        parent = self._base
        while parent:
            parent = parent.getparent()
        return _Element(parent)

    def get_parent(self) -> "_Element":
        """Retrieves the parent of this element if it exists.

        Returns:
            _Element: The parent element wrapper if available, otherwise None.
        """
        if self.is_literal:
            return None
        parent = self._base.getparent()
        if parent is None:
            return None
        return _Element(parent)

    def get_children(self) -> list["_Element"]:
        """Retrieves all direct children of this element.

        Returns:
            list[_Element]: A list of element wrappers for the children of this element.
        """
        return [_Element(child) for child in self._base.getchildren()]

    def get_attributes(self) -> dict[str, str]:
        """Retrieves all attributes of this element as a dictionary. This does not include special attributes (those prefixed with `@`) such as `@text`, `@tag`, etc.

        Returns:
            dict[str, str]: A dictionary containing attribute names and their corresponding values.
        """
        # TODO what about special attributes like @text, @tail or @prefix ?
        return dict(**self._base.attrib)

    def xpath(self, xpath: str, namespaces: dict[str, str]) -> list["_Element"]:
        """Evaluates an XPath expression from this element.

        Args:
            xpath (str): The XPath query string (see https://www.w3schools.com/xml/xpath_intro.asp for details on xpath queries)
            namespaces (dict[str, str]): A dictionary of namespace prefixes to XML URIs.

        Returns:
            list[_Element]: A list of elements matching the XPath query (empty if there was no match).
        """
        elements = self._base.xpath(xpath, namespaces=namespaces)
        if not isinstance(elements, list):
            elements = [elements]
        return [_Element(element) for element in elements]

    def index(self, element: "_Element") -> int:
        """Returns the index of the specified child element relative to this element.

        Args:
            element (_Element): The child element.

        Returns:
            int: The index of the child element.
        """
        return self._base.index(element._base)

    @property
    def prefix(self) -> str:
        """Retrieves the namespace prefix of this element, if any.

        Returns:
            str: The namespace prefix or the last segment of the default namespace URI if no explicit prefix is set.
        """
        prefix = self._base.prefix
        if prefix is None:
            # it is using the default namespace, fin
            prefix = self._base.nsmap[None].rsplit("/", 1)[-1]
        return prefix

    @property
    def tag(self) -> str:
        """Retrieves the tag of the element, excluding its namespace.

        Returns:
            str: The tag name of the element.
        """
        return self._base.tag.rsplit("}", 1)[-1]

    @property
    def name(self) -> str:
        """The qualified name of the element, combining the prefix and the tag.

        Returns:
            str: The fully qualified name of the element.
        """
        return f"{self.prefix}:{self.tag}"

    @property
    def text(self) -> str:
        """Retrieves the text content of this element.

        Returns:
            str: The text content of the element.
        """
        return self._base.text

    @text.setter
    def text(self, value: Any) -> None:
        """Sets the text content of this element to a string representation of the value.

        Args:
            value (Any): The text content
        """
        if isinstance(value, Expr):
            value = value.eval(self)
        self._base.text = str(value)

    @property
    def tail(self) -> str:
        """Retrieves the tail text of this element (text following this element).

        Returns:
            str: The tail text of the element.
        """
        return self._base.tail

    @tail.setter
    def tail(self, value: Any):
        """Sets the tail text of this element.

        Args:
                    value (Any): The text to set as the tail.
        """
        if isinstance(value, Expr):
            value = value.eval(self)
        self._base.tail = str(value)

    @property
    def head(self) -> str:
        """Retrieves the head text (text immediately preceding this element).

        Returns:
            str: The head text of the element.
        """
        prev = self._base.getprevious()
        if prev is None:
            parent = self._base.getparent()
            return parent.text
        else:
            return prev.tail

    @head.setter
    def head(self, value: Any):
        """Sets the head text of this element's parent or previous sibling.

        Args:
                    value (Any): The text to set as the head.
        """
        if isinstance(value, Expr):
            value = value.eval(self)
        prev = self._base.getprevious()
        if prev is None:
            parent = self._base.getparent()
            parent.text = value
        else:
            prev.tail = value

    def get(
        self, key: str, default: Any = None
    ) -> str | int | bool | float | list | dict | tuple:
        """Retrieves the value of an attribute or returns a default value if the attribute does not exist.

        Args:
            key (str): The attribute name.
            default (Any): The default value to return if the attribute is not found.

        Returns:
            Any: The value of the attribute, evaluated to a Python literal if possible, or the default value.
        """
        return _Element.literal_eval(self._base.get(key, default=default))

    def set(self, key: str, value: Any):
        """Sets the value of an attribute.

        Args:
            key (str): The attribute name.
            value (Any): The new value for the attribute.
        """
        if isinstance(value, Expr):
            value = value.eval(self)
        return self._base.set(key, str(value))

    def replace(self, old_element: "_Element", new_element: "_Element"):
        """Replaces an old child element with a new element.

        Args:
            old_element (_Element): The element to be replaced.
            new_element (_Element): The new element.
        """
        return self._base.replace(old_element._base, new_element._base)

    def insert(self, index: int, element: "_Element"):
        """Inserts an element at a specified index among this element's children.

        Args:
            index (int): The index at which the new element should be inserted.
            element (_Element): The element to insert.
        """
        return self._base.insert(index, element._base)

    def remove(self, element: "_Element"):
        """Removes a specific child element from this element.

        Args:
            element (_Element): The element to remove.
        """
        return self._base.remove(element._base)

    def remove_attribute(self, attribute_name: str):
        """Removes an attribute from this element.

        This does not work with special attributes (prefixed with `@`). Use the corresponding methods:
        - `@text` : `remove_text`
        - `@tail` : `remove_tail`
        - `@head` : `remove_head`
        Other special attributes (`@prefix`, `@tag` `@name`) cannot be removed.

        Args:
            attribute_name (str): The name of the attribute to remove.
        """
        del self._base.attrib[attribute_name]

    def remove_text(self):
        """Removes the text content of this element (sets the text to the empty string)."""
        self._base.text = None

    def remove_tail(self):
        """Removes the tail text content of this element (sets the text to the empty string)."""
        self._base.tail = None

    def remove_head(self):
        """Removes the head text content of this element (sets the text to the empty string)."""
        # TODO implement removal of head text
        raise NotImplementedError("TODO implement removal of head text")

    def is_orphaned(self, root: "_Element"):
        """Determines if this element is orphaned (is not connected to a specified root element).

        Args:
            root (_Element): The root element to check against.

        Returns:
            bool: `True` if this element is orphaned relative to the provided root, `False` otherwise.
        """
        parent = self._base
        while parent is not None:
            parent = parent.getparent()
            if parent == root._base:
                return False
        return True

    def __hash__(self):
        return self._base.__hash__()

    def __eq__(self, other):
        return self._base.__eq__(other)

    def __str__(self):
        return str(self._base)

    @property
    def attribute_name(self):
        """Retrieves the name of the attribute if this element is an attribute.

        Returns:
            str: The attribute name.
        """
        return self._base.attrname

    @property
    def is_attribute(self):
        """Checks if this element is an attribute node.

        Returns:
            bool: True if this element is an attribute node, otherwise False.
        """
        return self._base.is_attribute

    @property
    def is_head(self):
        """Checks if this element represents head text.

        Returns:
            bool: True if this element represents head text, otherwise False.
        """
        # TODO however, generally we should prefer to use is_text or is_tail and this check is more expensive...
        raise NotImplementedError("TODO check if this is a head")

    @property
    def is_text(self):
        """Checks if this element represents text content.

        Returns:
            bool: True if this element represents text content, otherwise False.
        """
        return self._base.is_text

    @property
    def is_tail(self):
        """Checks if this element represents tail text.

        Returns:
            bool: True if this element represents tail text, otherwise False.
        """
        return self._base.is_tail

    @property
    def is_literal(self):
        """Checks if this element is a literal value (int, float, or bool).

        Returns:
            bool: True if this element is a literal value, otherwise False.
        """
        return isinstance(self._base, int | float | bool)

    @property
    def is_unicode_result(self):
        """Checks if this element is a unicode result (this typically means the element is a text element, or is some other `str` result from an xpath query).

        Returns:
            bool: True if the element is an unicode result (see `lxml` docs for more details), otherwise False.
        """
        return isinstance(self._base, ET._ElementUnicodeResult)

    @property
    def is_element(self):
        """Checks if this element is a full xml element (i.e. is fully enclosed with `</>`).

        Returns:
            bool: True if the element represents a full xml element, otherwise False.
        """
        return isinstance(self._base, ET._Element)

    @property
    def nsmap(self):
        """Retrieves the namespace mapping for the element (if it is a full element).

        Returns:
            dict: A dictionary representing the namespace prefix to URI mapping.
        """
        return self._base.nsmap

    def as_string(self) -> str:
        """Converts the element to a canonical string representation.

        Returns:
                    str: A canonical string representation of the element.
        """
        return ET.tostring(
            self._base,
            method="c14n",
        ).decode("UTF-8")

    def as_literal(self):
        """Converts this element to a Python literal (only valid for unicode elements or attributes).

        Returns:
                    Any: The Python literal value of the element.

        Raises:
                    XMLQueryError: If the element cannot be converted to a literal.
        """
        if self.is_literal:
            return self._base
        elif self.is_unicode_result:
            return _Element.literal_eval(self._base)
        else:
            raise XMLQueryError(f"Failed to convert element {self} to literal.")

    def iter_parents(self):
        """Generator that traverses upward through the element tree yielding each parent element.

        Yields:
            _Element: The next parent element.
        """
        parent = self._base
        while parent:
            parent = parent.getparent()
            yield _Element(parent)

    @staticmethod
    def literal_eval(
        value: str,
    ) -> str | int | float | bool | list | tuple | dict | None:
        """Safely evaluates a string to a Python literal (str, int, float, bool, list, tuple, dict) if possible.

        Args:
            value (str): The string to evaluate.

        Returns:
            str | int | float | bool | list | tuple | dict | None: The evaluated Python literal, or the original string if evaluation fails - this means it could not be evalated to a python literal. Will return `None` if `None` is provided.
        """
        if value is None:
            return None
        try:
            return ast.literal_eval(value)
        except (SyntaxError, ValueError):
            return str(value)



"""Package defining the `XMLState` class along with its default implementation (based on `lxml`)."""

from abc import ABC, abstractmethod
from typing import Any
from functools import wraps
from lxml import etree as ET

from .query import Select, Update, Delete, Replace, Insert, XMLQueryError
from ._element import _Element, XML_START_PATTERN

__all__ = ("XMLState", "_XMLState")

# special variable keys
TEXT = "@text"  # inner text of the element
TAIL = "@tail"  # text that appears AFTER the element
HEAD = "@head"  # text that appears BEFORE the element
TAG = "@tag"  # <svg:g/> tag = "g"
NAME = "@name"  # <svg:g/> name = "svg:g"
PREFIX = "@prefix"  # <svg:g/> prefix = "svg"


def _set_xpath_on_exception(fun):
    """Utility decorator that sets the `xpath` attribute of an `XMLQueryError` if it is raised in a function."""

    @wraps(fun)
    def _set_xpath_on_exception(*args):
        try:
            return fun(*args)
        except XMLQueryError as e:
            e.kwargs["xpath"] = args[1].xpath
            raise

    return _set_xpath_on_exception


class XMLState(ABC):
    """A class to represent and manipulate XML data."""

    @abstractmethod
    def update(self, query: Update):
        """Updates elements in the XML state based on the provided `Update` query.

        Args:
            query (Update): update query
        """

    @abstractmethod
    def insert(self, query: Insert):
        """Inserts new elements into the XML state based on the provided `Insert` query. See the query class for details.

        Args:
            query (Insert): insert query
        """

    @abstractmethod
    def replace(self, query: Replace):
        """Replaces elements in the XML state based on the provided `Replace` query. See the query class for details.

        Args:
            query (Replace): replace query
        """

    @abstractmethod
    def delete(self, query: Delete):
        """Deletes elements from the XML state based on the provided `Delete` query. See the query class for details.

        Args:
            query (Delete): delete query
        """

    @abstractmethod
    def select(self, query: Select):
        """Retrieves data from the XML state based on the provided `Select` query. See the query class for details.

        Args:
            query (Select): select query
        """


class _XMLState(XMLState):
    """Default implementation of `XMLState`. Underlying xml parsing and queries are handled by the `lxml` package."""

    def __init__(
        self,
        xml: str,
        namespaces: dict[str, str] | None = None,
        parser: ET.XMLParser | None = None,
    ):
        super().__init__()
        if parser is None:
            parser = ET.XMLParser(remove_comments=True)
        self._parser = parser
        self._root = _Element(ET.fromstring(xml, parser=self._parser))
        self._namespaces = dict() if namespaces is None else namespaces

    def __str__(self):
        return str(ET.tostring(self._root._base, method="c14n2", with_comments=False))

    def xpath(self, xpath: str) -> list[_Element]:
        """Query inner xml using xpath producing a (possibly empty) list of elements that are the result of the query.

        Args:
            xpath (str): xpath query

        Returns:
            list[_Element]: elements that result from the query
        """
        return self._root.xpath(xpath, namespaces=self._namespaces)

    def get_root(self) -> _Element:
        """Get the root element.

        Returns:
            _Element: the root element.
        """
        return self._root

    def get_namespaces(self) -> dict[str, str]:
        """Get XML namespaces (prefix -> URI).

        Returns:
            dict[str, str]: namespaces
        """
        return self._namespaces

    @_set_xpath_on_exception
    def update(self, query: Update) -> None:
        """Updates the attributes of XML element(s) based on the `Update` query.

        Args:
            query (Update): query
        """
        elements = self.xpath(query.xpath)
        for element in elements:
            _XMLState.update_element_attributes(element, query.attrs)

    @_set_xpath_on_exception
    def insert(self, query: Insert) -> None:
        """Inserts an XML element based on the `Insert` query.

        Args:
            query (Insert): query

        Raises:
            XMLQueryError: if a parent element could not be found (this is defined by the `xpath` of the Insert query)
            XMLQueryError: If multiple parents were found - this is not currently supported by may be in the future.
        """
        elements = self.xpath(query.xpath)
        if len(elements) == 0:
            raise XMLQueryError(
                "Invalid xpath: `{xpath}` for `insert`, no parent element was found at this path.",
            )
        if len(elements) > 1:
            raise XMLQueryError(
                "Invalid xpath: `{xpath}` for `insert`, found {elements_length} but only one is allowed.",
                elements_length=len(elements),
            )
        return _XMLState.insert_in_element(
            elements[0],
            query,
            parser=self._parser,
        )

    @_set_xpath_on_exception  # TODO implement Replace!
    def replace(self, query: Replace) -> None:
        """Replaces an XML element based on the `Replace` query.

        NOTE: THIS QUERY TYPE IS CURRENTLY NOT IMPLEMENTED but is planned as part of the next release.

        Args:
            query (Replace): query

        Raises:
            XMLQueryError: If multiple elements were found to replace (only one is allowed).
            NotImplementedError: THIS QUERY TYPE IS CURRENTLY NOT IMPLEMENTED
        """
        elements = self.xpath(query.xpath)
        if len(elements) > 1:
            raise XMLQueryError(
                "Invalid xpath: `{xpath}` for `replace`, found {elements_length} but only one is allowed.",
                elements_length=len(elements),
            )
        raise NotImplementedError(
            "TODO element replacement (`replace`) is not currently implement."
        )

    @_set_xpath_on_exception
    def delete(self, query: Delete) -> None:
        """Deletes one or more XML elements based on the `Delete` query.

        Args:
            query (Delete): query

        Raises:
            XMLQueryError: If no elements were found for deletion.
        """
        elements = self.xpath(query.xpath)
        if len(elements) == 0:
            raise XMLQueryError(
                "No elements was found for deletion at xpath: {xpath}",
            )
        for element in elements:
            _XMLState.delete_element(element)

    @_set_xpath_on_exception
    def select(self, query: Select) -> list[Any]:
        """Select an element or its attributes based on the `Select` query.

        Args:
            query (Select): query

        Returns:
            list[Any]: list of results of the select (one per xpath result), typically will consist of python literal types (int, float, bool, str, list, dict).
        """
        elements = self.xpath(query.xpath)
        result = [_XMLState.select_from_element(element, query) for element in elements]
        # print(result)
        return result

    @staticmethod
    def update_element_attributes(element: _Element, attrs: dict[str, Any]):
        if not element.is_element:
            raise XMLQueryError(
                "Failed to update: `{element}` is not an xml element. (xpath: `{xpath}`)",
                element=element,
            )
        for attr, value in attrs.items():
            if not attr.startswith("@"):
                element.set(attr, value)
            elif attr == TEXT:
                element.text = value
            elif attr == TAIL:
                element.tail = value
            elif attr == HEAD:
                element.head = value
            elif attr == TAG:
                raise XMLQueryError(
                    "Cannot update tag on an element. (xpath: `{xpath}`)"
                )
            elif attr == PREFIX:
                raise XMLQueryError(
                    "Cannot update namespace prefix on an element. (xpath: `{xpath}`)"
                )

    @staticmethod
    def _replace_element(
        element: _Element,
        xml: str,
        parser: ET.XMLParser,
    ):
        parent = element.get_parent()
        if parent is None:
            raise NotImplementedError(
                "Replacing the XML root node is not yet supported."
            )
        replace = _XMLState._new_element(
            xml,
            parser=parser,
        )
        parent.replace(element, replace)

    @staticmethod
    def _update_unicode_element(element: _Element, value: Any):
        parent = element.get_parent()
        if element.is_attribute:
            parent.set(element.attribute_name, value)
        elif element.is_text:
            parent.text = value
        elif element.is_tail:
            parent.tail = value
        else:
            # NOTE who knows what happened if this occurs, if it does more extensive testing is needed...
            raise XMLQueryError(
                "Failed to update unicode element: `{element}` unknown element type. (xpath: `{xpath}`)",
                element=element,
            )

    @staticmethod
    def insert_in_element(
        element: _Element,
        query: Insert,
        parser: ET.XMLParser,
        # inherit_namespaces: bool,
    ):
        if element.is_element:
            if XML_START_PATTERN.match(query.element):
                child = _XMLState._new_element(query.element, parser=parser)
                element.insert(query.index, child)
            else:
                _XMLState._insert_text_at(element, query.element, index=query.index)
        else:
            raise XMLQueryError(
                "Failed to insert into xpath result: `{element}` must be an xml element. (xpath: `{xpath}`)",
                element=element,
            )

    @staticmethod
    def _insert_text_at(element: _Element, text: str, index: int):
        if index > 0:
            children = element.get_children()
            children[index - 1].tail = text
        else:
            element.text = text

    @staticmethod
    def _new_element(
        xml: str,
        parser: ET.XMLParser,
    ) -> _Element:
        # TODO implement this inside _Element, we want to avoid using ET everywhere in this class
        return _Element(ET.fromstring(xml, parser=parser))

    @staticmethod
    def delete_element(element: _Element):
        if element.is_literal:
            raise XMLQueryError(
                "Failed to delete element: `{element}` as it is a literal. (xpath: `{xpath}`)",
                element=element,
            )
        elif element.is_element:
            element.get_parent().remove(element)
        elif element.is_attribute:
            element.get_parent().remove_attribute(element.attribute_name)
        elif element.is_text:
            element.get_parent().remove_text()
        elif element.is_tail:
            element.get_parent().remove_tail()
        else:
            # who knows what happened if this occurs
            raise XMLQueryError(
                "Failed to delete xpath result: `{element}` unknown element type (xpath: `{xpath}`)",
                element=element,
            )

    @staticmethod
    def _delete_element_attributes(element: _Element, attrs: list[str]):
        for attr in attrs:
            if not attr.startswith("@"):
                element.remove_attribute(attr)
            elif attr == TEXT:
                element.remove_text()
            elif attr == TAIL:
                element.remove_tail()
            elif attr == HEAD:
                # TODO
                raise NotImplementedError(
                    "Removing the head is not implemented yet use @text or @tail for now."
                )
            elif attr == TAG:
                # this is not a valid attribute to delete, all elements must have a name.
                raise XMLQueryError(
                    "Cannot delete tag from an element. (xpath: `{xpath}`)"
                )
            elif attr == PREFIX:
                # this is not a valid attribute to delete, all elements must have a name.
                raise XMLQueryError(
                    "Cannot delete namespace prefix from an element. (xpath: `{xpath}`)"
                )

    @staticmethod
    def select_from_element(element: _Element, query: Select):
        if element.is_element:
            if query.attrs:
                return dict(_XMLState._iter_element_attributes(element, query.attrs))
            else:
                return element.as_string()
        elif element.is_unicode_result:
            if query.attrs:
                raise XMLQueryError(
                    "Failed to select attributes: `{element}` is not an xml element, xpath: `{xpath}`",
                    element=element,
                )
            else:
                return element.as_literal()
        elif element.is_literal:
            return element.as_literal()  # this is already a primitive value
        else:
            # TODO could be a string or int value if an attribute is select? test this
            raise XMLQueryError(
                "Failed to select xpath result: `{element}`, xpath: `{xpath}`",
                element=element,
            )

    @staticmethod
    def _iter_element_attributes(element: _Element, attrs: list[str]):
        """Get all attribute values from an element.

        Args:
            element (_Element): element to get attribute values from.
            attrs (list[str]): attributes to get

        Raises:
            XMLQueryError: if an unknown special attribute is provided as part of `attrs`.

        Yields:
            Tuple[str, Any] : the attribute name and value as (attr, value).
        """
        for attr in attrs:
            if not attr.startswith("@"):
                yield attr, element.get(attr, None)
            elif attr == TAG:
                yield attr, element.tag
            elif attr == NAME:
                yield attr, element.name
            elif attr == PREFIX:
                yield attr, element.prefix
            elif attr == TEXT:
                yield attr, element.text
            elif attr == TAIL:
                yield attr, element.tail
            elif attr == HEAD:
                yield attr, element.head
            else:
                raise XMLQueryError(
                    f"Unknown special attribute: {attr}, must be one of: {TAG, NAME, PREFIX, TEXT, TAIL, HEAD}"
                )



"""Contains the default `Ambient` (see `star_ray`) implementation that uses XML as its state description language and xpath as its query language."""

from typing import Any
from star_ray import Ambient, Agent
from star_ray.event import ActiveObservation, ErrorActiveObservation
from star_ray.pubsub import Subscribe, Unsubscribe

from .state import XMLState, _XMLState
from .query import Select, XMLQuery

DEFAULT_XML = "<xml></xml>"
DEFAULT_NAMESPACES = {}


class XMLAmbient(Ambient):
    """An implementation of an `Ambient` (see `star_ray`) that uses XML as its state description language and xpath as its query language."""

    def __init__(
        self,
        agents: list[Agent],
        xml: str | None = None,
        namespaces: dict[str, str] | None = None,
        xml_state: XMLState | None = None,
        **kwargs: dict[str, Any],
    ):
        """Constructor.

        Args:
            agents (list[Agent]): list of agents to add to this `Ambient` initially.
            xml (str | None, optional): initial xml data. Defaults to <xml></xml>.
            namespaces (dict[str, str], optional): namespace map associated with the initial `xml` data. Defaults to an empty dict.
            xml_state (XMLState | None, optional): XMLState to use as the underlying state. Defaults to using `star_ray_xml._XMLState` with the arguments `xml` and `namespaces` as provided.
            kwargs (dict[str, Any]): Additional optional arguments.
        """
        super().__init__(agents)
        self._state = None
        if xml_state is None:
            self._state = _XMLState(
                xml if xml else DEFAULT_XML,
                namespaces=namespaces if namespaces else DEFAULT_NAMESPACES,
            )
        else:
            assert xml is None  # set these directly on the `xml_state`
            assert namespaces is None  # set these directly on the `xml_state`
            self._state = xml_state

    def get_state(self) -> XMLState:
        """Get the underlying `XMLState`, this should be read only and NEVER modified without a call to `__update__` to prevent unexpected issues.

        Returns:
            XMLState: the current state of this `Ambient`.
        """
        return self._state  # NOTE: this is read only!

    def __select__(
        self, action: XMLQuery | Subscribe | Unsubscribe
    ) -> ActiveObservation | ErrorActiveObservation:
        """Execute a read-only action in this `Ambient`. These actions must derive `XMLQuery` or be a subscription action to avoid an unknown action error.

        Args:
            action (XMLQuery | Subscribe | Unsubscribe): action to execute

        Raises:
            ValueError: if the action type is unknown.

        Returns:
            ActiveObservation | ErrorActiveObservation: the resulting observation.
        """
        try:
            if isinstance(action, XMLQuery) and action.is_read:
                values = action.__execute__(self._state)
                if (
                    values is not None
                ):  # TODO typically the result wont be None... perhaps something has gone wrong if it does?
                    return ActiveObservation(action_id=action, values=values)
            elif isinstance(action, Subscribe | Unsubscribe):
                return self.__subscribe__(action)
            else:
                raise ValueError(
                    f"{action} does not derive from one of required type(s):`{[Select, Subscribe, Unsubscribe]}`"
                )
        except Exception as e:
            return ErrorActiveObservation.from_exception(action=action, exception=e)

    def __update__(
        self, action: XMLQuery
    ) -> ActiveObservation | ErrorActiveObservation | None:
        """Execute a write action in this `Ambient`. These actions must derive `XMLQuery` to avoid an unknown action error.

        Args:
            action (XMLQuery): the action to execute.

        Returns:
            ActiveObservation | ErrorActiveObservation | None: the resulting observation
        """
        try:
            values = action.__execute__(self._state)
            if values is not None:
                return ActiveObservation(action_id=action, values=values)
        except Exception as e:
            return ErrorActiveObservation.from_exception(action, e)

    def __subscribe__(  # TODO perhaps this should be supported... why isn't it?
        self, action: Subscribe | Unsubscribe
    ) -> ActiveObservation | ErrorActiveObservation:
        """Subscribe to receive events from this ambient. THIS IS NOT SUPPORTED by `XMLAmbient`. If you wish to subscribe to receive XML related events subclass `XMLAmbient` and implement publishing of actions.

        Args:
            action (Subscribe | Unsubscribe): action to execute

        Raises:
            ValueError: if this method is called, it is not a supported operation.

        Returns:
            ActiveObservation | ErrorActiveObservation: _description_
        """
        try:
            raise ValueError(
                f"`star_ray` pub-sub is not supported by ambient of type: `{XMLAmbient}`."
            )
        except Exception as e:
            return ErrorActiveObservation.from_exception(action, e)



"""Package defining XML query classes.

These include:
- The `XMLQuery` class, which should be the base class for all XML queries.
- The primitive XML queries: `Select`, `Update`, `Insert`, `Delete`, `Replace`.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING
from pydantic import BaseModel
from star_ray.event import Action
from star_ray.utils.literal_eval import literal_eval_with_ops

if TYPE_CHECKING:
    from .state import XMLState, _Element

__all__ = ("XMLQuery", "XPathQuery", "Update")


class _format_dict_template(dict):
    def __missing__(self, key):
        return f"{{{key}}}"


class Expr(BaseModel):  # TODO test this
    """This class represents a simple expression that can be evaluated as part of an XML query.

    It uses `star_rays`'s `literal_eval_with_ops` function under the hood to evaluate an expression which follows python syntax. It can peform simple arithmetic operations on attributes of an element (which are resolved during execution).

    Variables are defined in single `{` or `}` as in pythons string formatting. Example expression: `Expr("{x} + {y}")` assuming `x` and `y` are attributes of the element that will be queried.

    Example:
        `Expr` can be used in an `Update` query:
        ```
           Update(xpath="/svg:rect", attrs={"width" : Expr("{width} + 10")})
        ```

    Note that special attributes (prefixed with `@`) are not currently supported.
    """

    expr: str

    def __init__(self, expr: str, **values: dict[str, Any]):
        expr = expr.format_map(_format_dict_template(values))
        super().__init__(expr=expr)

    def eval(self, element: _Element) -> Any:
        """Evaluate this expression given the element as context.

        Args:
            element (_Element): element to use as context.

        Returns:
            Any: the result of the evaluation, which is typically a python literal (e.g. int, float, bool, str, list, dict).
        """
        expr = self.expr.format_map(element.get_attributes())
        result = literal_eval_with_ops(expr)
        return result


class XMLQueryError(Exception):
    """Error that may occur during the execution of an `XMLQuery`."""

    def __init__(self, message, **kwargs):
        super().__init__(message)
        self.kwargs = kwargs

    def __str__(self):
        return self.args[0].format(**self.kwargs)

    def __repr__(self):
        return str(self)


class XMLQuery(ABC, Action):
    """Base class for XML queries. Defines the `__execute__` api.

    `__execute__` will be called during the execution of this action in the `Ambient`. It is effectively the definition of this action.
    """

    @abstractmethod
    def __execute__(self, state: XMLState) -> Any:
        """Execute method that defines this action. In this method you have direct access to the `XMLState` and can read/write properties by using the avaliable methods. Typically this will involve executing more primitive XMLQueries (such as `Select` and `Update`).

        Args:
            state (XMLState): the current state

        Returns:
            Any: The result of executing this query. For write operations this may be None. Typically for read operations this will be some xml data or derived data.
        """
        pass

    @property
    @abstractmethod
    def is_read(self):
        """Whether this action is READ ONLY."""
        pass

    @property
    @abstractmethod
    def is_write(self):
        """Whether this action mutates the XML state."""
        pass

    @property
    @abstractmethod
    def is_write_tree(self):
        """Whether this action mutates the XML tree (e.g. insert, replace or delete elements) or will mutate multiple elements simultaneously."""
        pass

    @property
    @abstractmethod
    def is_write_element(self):
        """Whether this action mutates XML attributes on a SINGLE ELEMENT. This is a stronger guarantee than `is_write_tree` and may mean that the action can be executed in parallel."""
        pass


class XPathQuery(XMLQuery):
    """Base class for XML queries that use `xpath`."""

    xpath: str


class XMLUpdateQuery(XMLQuery):
    """Base class that defines properties: `is_read`, `is_write`, `is_write_element`, `is_write_tree`, for events that will only ever write to elements attributes."""

    @property
    def is_read(self):  # noqa: D102
        return False

    @property
    def is_write(self):  # noqa: D102
        return True

    @property
    def is_write_tree(self):  # noqa: D102
        return False

    @property
    def is_write_element(self):  # noqa: D102
        return True


class Update(XPathQuery, XMLUpdateQuery):
    """Query to update XML element attributes."""

    attrs: dict[str, int | float | bool | str | Expr | Expr]

    @staticmethod
    def new(xpath: str, attrs: dict[str, Any]):
        """Factory method for `Update` with positional arguments.

        Args:
            xpath (str): xpath used to locate the element.
            attrs (dict[str, Any]): attributes to update.

        Returns:
            Update: update query.

        See:
            `update` for further details.
        """
        return Update(xpath=xpath, attrs=attrs)

    def __execute__(self, state: XMLState) -> Any:  # noqa: D105
        return state.update(self)


class Insert(XPathQuery):
    """Query to insert an XML element."""

    element: str
    index: int

    @staticmethod
    def new(xpath: str, element: str, index: int = 0):
        """Factory method for `Insert` with positional arguments.

        GOTCHA: the inserted `element` must contain all relevant namespace information and its tag (or name) should be qualified with a prefix. For example: `<svg:rect xmlns:svg="http://www.w3.org/2000/svg" .../>`. If this is not done then the element may not be properly resolved by future XML queries.

        Args:
            xpath (str): xpath used to locate the parent element where the given `element` will be inserted.
            element (str): element to insert.
            index (int): the index to insert at.

        Returns:
            Insert: insert query.

        See:
            `insert` for further details.
        """
        return Insert(
            xpath=xpath,
            element=element,
            index=index,
        )

    @property
    def is_read(self):
        return False

    @property
    def is_write(self):
        return True

    @property
    def is_write_tree(self):
        return True

    @property
    def is_write_element(self):
        return False

    def __execute__(self, state: XMLState) -> Any:
        return state.insert(self)


class Delete(XPathQuery):
    """Query to delete an XML element."""

    @staticmethod
    def new(xpath: str):
        """Factory method for `Delete` with positional arguments.

        Args:
            xpath (str): the xpath of the element(s) to delete.

        Returns:
            Delete: the delete query.
        See:
            `delete` for further details.
        """
        return Delete(xpath=xpath)

    @property
    def is_read(self):
        return False

    @property
    def is_write(self):
        return True

    @property
    def is_write_tree(self):
        return True

    @property
    def is_write_element(self):
        return False

    def __execute__(self, state: XMLState) -> Any:
        return state.delete(self)


class Replace(XPathQuery):
    """Query to replace an XML element."""

    element: str

    @staticmethod
    def new(xpath: str, element: str):
        """Factory method for `Replace` with positional arguments.

        GOTCHA: the `element` to be the replacement must contain all relevant namespace information and its tag (or name) should be qualified with a prefix. For example: `<svg:rect xmlns:svg="http://www.w3.org/2000/svg" .../>`. If this is not done then the element may not be properly resolved by future XML queries.

        Args:
            xpath (str): the xpath of the element(s) to replace, the result of which must be an XML element.
            element (str): element to replace with.

        Returns:
            Replace: the replace query.

        See:
            `replace` for further details.
        """
        return Replace(xpath=xpath, element=element)

    @property
    def is_read(self):
        return False

    @property
    def is_write(self):
        return True

    @property
    def is_write_tree(self):
        return True

    @property
    def is_write_element(self):
        return False

    def __execute__(self, state: XMLState) -> Any:
        return state.replace(self)


class Select(XPathQuery):
    """Query to select XML elements and their attributes."""

    attrs: list[str] | None

    @staticmethod
    def new(xpath: str, attrs: list[str] = None):
        """Factory method for `Select` with positional arguments.

        Args:
            xpath (str): the xpath of the element(s) to select.
            attrs (list[str], optional): attributes to select. Defaults to None, which will cause the entire element to be selected.

        Returns:
            Select: the select query.

        See:
            `select` for further details.
        """
        return Select(xpath=xpath, attrs=attrs)

    @property
    def is_read(self):
        return True

    @property
    def is_write(self):
        return False

    @property
    def is_write_tree(self):
        return False

    @property
    def is_write_element(self):
        return False

    def __execute__(self, state: XMLState) -> Any:
        return state.select(self)


def insert(xpath: str, element: str, index: int = 0):
    """TODO."""
    return Insert(xpath=xpath, element=element, index=index)


def delete(xpath: str):
    """TODO."""
    return Delete(xpath=xpath)


def replace(xpath: str, element: str):
    """TODO."""
    return Replace(xpath=xpath, element=element)


def update(xpath: str, attrs: dict[str, Any]):
    """Update XML data.

    Args:
        xpath (str): xpath used to locate the element.
        attrs (dict[str, Any]): attributes to update.

    XML elements hold different kinds of data which can be updated as follows:
    - attributes      : `<attribute_name>` e.g. `id` or `width`
    - text            : `@text` this text appears INSIDE the element e.g. `<g>text</g>`
    - tail (text)     : `@tail` this text appears AFTER the element e.g. `<g></g>text`
    - head (text)     : `@head` this text appears BEFORE the element e.g. `text<g></g>`

    A note about updating text:
    An element `<g></g>` contains no text. Its text can be updated using the `@text` attribute, but it cannot be updated using an xpath containing `/text()'.
    An element `<g> </g>` contains the text " ", and can therefore be updated using both methods.

    There is currently no way to update the tag (element name) or namespace information of an element, this is due to an unfortunate limitation with `lxml`.
    """
    return Update(xpath=xpath, attrs=attrs)


def select(xpath: str, attrs: list[str] = None):
    """Select XML data.

    Args:
        xpath (str): the xpath of the element(s) to select.
        attrs (list[str], optional): attributes to select. Defaults to None, which will cause the entire element to be selected.

    XML elements hold different kinds of data which can be selected as follows:
    - tag             : `@tag`
    - prefix          : `@prefix` this is the namespace prefix e.g. `svg` in `svg:rect`
    - attributes      : `<attribute_name>` e.g. `id` or `width`
    - text            : `@text` this text appears INSIDE the element e.g. `<g>text<g>`
    - tail (text)     : `@tail` this text appears AFTER the element e.g. `<g><g>text`
    - child elements  : are select using the xpath query, e.g. //xml/rect/*, attrs should be `None`

    There are alternative ways to get the tag and text attributes using xpath directly (attrs should be `None`):
    - tag             : name(QUERY)  this will return the tag
    - text            : QUERY/text() this will return all string text in the element as a list (if it exists)

    All of the other built-in xpath functions (`count(...)` `sum(...)`, etc.) can be used to get element properties.

    It is also possible to get a single attribute using xpath directly (attrs should be `None`):
        xpath = "//svg:rect/@width
    by using `attrs` one can select multiple attributes with a single query.

    A known querk of selecting with /text():
        xpath = "//svg:g/text()"
        <g> </g> will return [" "]
        <g></g>  will return [] instead of what we might expect [""]

    Returns:
        Select: select query
    """
    return Select(xpath=xpath, attrs=attrs)


# TODO do the others is_select_query, is_insert_query, is_replace_query, is_delete_query


def is_update_query(cls):
    """Class decorator that will automatically add relevant properties is_read, is_write, is_write_element, is_write_tree to a class treating it as an Update query."""

    @property
    def is_read(self):
        return False

    @property
    def is_write(self):
        return True

    @property
    def is_write_tree(self):
        return False

    @property
    def is_write_element(self):
        return True

    # Add the properties to the class
    cls.is_read = is_read
    cls.is_write = is_write
    cls.is_write_tree = is_write_tree
    cls.is_write_element = is_write_element

    return cls
